from typing import Any, Dict, List

from pydantic import BaseModel, Field, model_validator

from promptquality.types.custom_scorer import CustomMetricType, CustomScorer
from promptquality.types.rows import PromptRow
from promptquality.utils.logger import logger


class UserSubmittedMetrics(BaseModel):
    scorer_name: str = "_user_submitted"
    name: str = Field(serialization_alias="metric_name")

    scores: List[CustomMetricType] = Field(default_factory=list)
    indices: List[int] = Field(default_factory=list)

    aggregates: Dict[str, CustomMetricType] = Field(default_factory=dict)

    @model_validator(mode="before")
    def validate_scores_and_indices(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if len(values["scores"]) != len(values["indices"]):
            raise ValueError("Length of scores must match length of indices.")
        return values

    @classmethod
    def from_scorer(cls, scorer: CustomScorer, prompt_rows: List[PromptRow]) -> "UserSubmittedMetrics":
        scores, indices = [], []
        for i, row in enumerate(prompt_rows):
            try:
                score = scorer.scorer_fn(row)
                scores.append(score)
                indices.append(row.index)
            except Exception as exception:
                logger.warning(
                    f"Failed to score response: {row.response} at index {i}, exception:"
                    f"{exception}. Skipping row {i}."
                )
        if scorer.aggregator_fn:
            aggregates = scorer.aggregator_fn(scores, indices)
        else:
            logger.debug(f"No aggregator set for scorer {scorer.name}.")
            aggregates = dict()
        return cls(name=scorer.name, scores=scores, indices=indices, aggregates=aggregates)



from types import SimpleNamespace

ConfigEnvironmentVariables = SimpleNamespace(
    console_url="GALILEO_CONSOLE_URL",
    username="GALILEO_USERNAME",
    password="GALILEO_PASSWORD",
    api_key="GALILEO_API_KEY",
)



from os import getenv
from typing import Optional

from promptquality.constants.config import ConfigEnvironmentVariables
from promptquality.types.config import Config
from promptquality.utils.config import get_config_location


def set_config(console_url: Optional[str] = None) -> Config:
    """
    Set the config for `promptquality`.

    If the config file exists, and console_url is not passed, read it and return the
    config. Otherwise, set the default console URL and return the config.

    Parameters
    ----------
    console_url : Optional[str], optional
        URL to the Galileo console, by default None and we use the Galileo Cloud URL.

    Returns
    -------
    Config
        Config object for `promptquality`.
    """

    if not console_url and get_config_location().exists():
        config = Config.read()
    else:
        console_url = console_url or getenv(ConfigEnvironmentVariables.console_url)
        if not console_url:
            raise ValueError(
                "Please pass your Galileo console URL, or set the environment "
                "variable `GALILEO_CONSOLE_URL` to the Galileo Console URL."
            )
        config = Config(console_url=console_url)
    config.write()
    return config



from typing import Optional

from pydantic import UUID4

from promptquality.helpers import get_project, get_project_from_name, get_templates
from promptquality.types.run import BaseTemplateResponse


def get_template(
    project_name: Optional[str] = None, project_id: Optional[UUID4] = None, template_name: Optional[str] = None
) -> BaseTemplateResponse:
    """
    Get a template for a specific project.

    Parameters
    ----------
    project_name : Optional[str]
        Project name.
    project_id : Optional[UUID4]
        Project ID.
    template_name : Optional[str]
        Template name.

    Returns
    -------
    BaseTemplateResponse
        Template response.
    """
    if project_name:
        project = get_project_from_name(project_name=project_name)
    elif project_id:
        project = get_project(project_id)
    else:
        raise Exception("Either project_id or project_name must be provided.")

    if not project:
        raise ValueError("Project not found.")

    if not template_name:
        raise Exception("template_name must be provided.")

    templates = get_templates(project.id)
    requested_template = None
    for template in templates:
        if template.name == template_name:
            requested_template = template
    if not requested_template:
        raise Exception(f"Template {template_name} not found.")
    return requested_template



from random import random
from time import sleep
from typing import Optional

from pydantic import UUID4
from tqdm.auto import tqdm

from promptquality.constants.job import JobStatus
from promptquality.constants.scorers import Scorers
from promptquality.helpers import get_job_status, get_run_scorer_jobs
from promptquality.types.config import Config
from promptquality.utils.logger import logger


def job_progress(job_id: Optional[UUID4] = None, config: Optional[Config] = None) -> UUID4:
    backoff = random()
    job_status = get_job_status(job_id, config)
    if JobStatus.is_incomplete(job_status.status):
        job_progress_bar = tqdm(
            total=job_status.steps_total,
            position=0,
            leave=True,
            desc=job_status.progress_message,
        )
        while JobStatus.is_incomplete(job_status.status):
            sleep(backoff)
            job_status = get_job_status(job_id, config)
            job_progress_bar.set_description(job_status.progress_message)
            job_progress_bar.update(job_status.steps_completed - job_progress_bar.n)
            backoff = random()
        job_progress_bar.close()
    logger.debug(f"Job {job_id} status: {job_status.status}.")
    if JobStatus.is_failed(job_status.status):
        raise ValueError(f"Job failed with error message {job_status.error_message}.") from None
    print("Initial job complete, executing scorers asynchronously. Current status:")
    scorer_jobs_status(config=config)
    return job_status.id


def scorer_jobs_status(
    project_id: Optional[UUID4] = None, run_id: Optional[UUID4] = None, config: Optional[Config] = None
) -> None:
    scorer_jobs = get_run_scorer_jobs(project_id, run_id)
    for job in scorer_jobs:
        if job.request_data is None or job.request_data.prompt_scorer_settings is None:
            logger.debug(f"Scorer job {job.id} has no scorer settings.")
            continue
        scorer_name = job.request_data.prompt_scorer_settings.scorer_name
        try:
            scorer_name = Scorers(scorer_name).name
        except ValueError:
            pass
        logger.debug(f"Scorer job {job.id} has scorer {scorer_name}.")
        if JobStatus.is_incomplete(job.status):
            print(f"{scorer_name.lstrip('_')}: Computing 🚧")
        elif JobStatus.is_failed(job.status):
            print(f"{scorer_name.lstrip('_')}: Failed ❌, error was: {job.error_message}")
        else:
            print(f"{scorer_name.lstrip('_')}: Done ✅")
